(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

(* $Id: abs-focimc.ml,v 1.21 2007/08/29 01:41:53 beyer Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This file contains all the code for the Abstraction module that pertains
   to focimodelchecker. The first half is for "concrete" focimodelchecker,
   the second half is for "abstract" focimodelchecker.
 *)

(************** CONCRETE FOCIMODELCHECKER   ***************************)




(***********************************************************************************)
(*focimodelchecker: foci_post. updates lvalue map, location info [CONCRETE] ***)
(************************************************************************************)

  let pred_to_data_region pred = Region.Concrete {Region.pred = pred}
(*	
  type lvmap_t = ((int * Expression.lval), Expression.expression) Hashtbl.t
*)
(* we need to know ALL interesting lvals of the program, for aliasing reasons *)
  let foci_load_lv_table () = () (* load all interesting lvals into the trace_lv_table *)

(* return new location and new stack *)
  let location_post atomic_reg op = 
    (* let (new_location, new_stack) = *)
    let callStack = Region.list_of_stack atomic_reg.Region.stack in  
    if (C_SD.isFunCall op) then
        begin
          match (C_SD.get_command op).C_SD.Command.code with
              C_SD.Command.FunctionCall fcexp -> 
                let (_fname,_,target) = C_SD.deconstructFunCall fcexp in
	
                  if (_fname = C_SD.__NotImplementedFunctionName || C_SD.is_noexpand _fname) 
                  then 
		    ((C_SD.get_target op), atomic_reg.Region.stack)
                  else   
		    begin
                      if ( C_SD.is_defined _fname) then
			begin
			  let src = C_SD.get_source op in
			    if List.mem src callStack then 
			      raise RecursionException
			    else
			      begin
				if (List.length callStack = !maxStackLen) 
				then (maxStackLen := !maxStackLen + 1;
				      M.msg_string M.Normal ("Max stack size: "^(string_of_int !maxStackLen)));
				let newStack = src:: callStack in
				  ((C_SD.lookup_entry_location _fname),
				      Region.CallStack newStack)
			      end
			end
                      else (C_SD.get_target op,atomic_reg.Region.stack) 
(* failwith ("Undeclared important function"^_fname)*)
		    end
	    | _ -> failwith "what the (post) !"
	end
      else 
        try 
	  if (C_SD.isReturn op) then 
	    begin
	      (*let _ = stackInfo := Some (tos_fname,tos_2_fname,true) in *)
	      let (ret, subs, nextLoc) = Region.reg_lastTarget (Region.Atomic atomic_reg) op in
		Absutil.set_ret_pointers ret subs (C_SD.get_location_fname nextLoc);
		let newStack =
		  List.tl callStack
		in
		  (nextLoc,(Region.CallStack newStack) ) 
	    end
	  else 
	    (* let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in *)
	      (C_SD.get_target op,atomic_reg.Region.stack) 
	with Failure ("tl") -> 
	  (* let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in *)
	    (C_SD.get_target op, atomic_reg.Region.stack ) 
	    


(* foci_post: (Region.t * lvmap_t ) -> Operation.t -> (Region.t * lvmap_t * P.predicate) *)
		      
  let foci_post (reg,orig_lvmap) op =
    let _ = M.msg_string M.Debug "Foci Post:" in
    let _ = M.msg_string M.Debug (Operation.to_string op) in
    (* 1. compute the pred, next lvmap *)
    let lvmap = lvmap_copy orig_lvmap in
    let cmd = C_SD.get_command op in
     (* 1-a. figure out if this is a return statement *) 
    let ret_o =
      if (C_SD.isReturn op) then
	try
	  let r = (Misc.fst3 (Region.reg_lastTarget reg op)) in
	    if r <> C_SD.junkRet then Some r else None
	with _ -> None (* reg_lastTarget can throw an exception if the stack is empty -- for "main" *)
      else None
    in
       (* RJ: only compute newpred, as the lvmap to be returned is just lvmap which will be updated *)
    let op_constraint =
     let rec _doBlock stmtlst pred i (* RJ: Whats this i for ? *) =
       match stmtlst with
	   [] -> pred
	 | stmt::rest -> 
	     begin
	       let sp_pred = block_sp lvmap ret_o  stmt pred in
		 _doBlock rest sp_pred (i+1)
	     end
     in
       match cmd.C_SD.Command.code with
	   C_SD.Command.Block el -> _doBlock el (P.True) 0
	 | C_SD.Command.Pred e -> 
	     begin 
	       let rv =  (* should take a boolean closure here -- see POPL04 paper *)
		 add_to_lv_table (Absutil.getVarsAndDerefs e); 
	     (* IRRELEVANT. everything of note should already have been added *)
	     (pred_lval_map_a lvmap) 0 e (* P.And (e::pred_l) *)
                in
      M.msg_string M.Debug ("bwp pred: "^(P.toString rv));
      rv
	     end
	 | C_SD.Command.FunctionCall fcallExp -> 
	     let (fname,_subs,target) = (C_SD.deconstructFunCall fcallExp) in
	     let exists_fun = 
	       try 
		 let fml = C_SD.lookup_formals fname 
		 in true 
	       with C_SD.NoSuchFunctionException _ -> false
	     in
	     let asgn_list = 
	       if (exists_fun || (target = C_SD.junkRet)) then
		 (List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x),e))) _subs )
		 @ (if (C_SD.is_noexpand fname) then [C_SD.Command.Expr fcallExp] else [])
	       else
		 [C_SD.Command.Expr( Expression.Assignment (Expression.Assign, target, Absutil.get_new_unknown fname ))]
	     in         
	       _doBlock asgn_list (P.True) 0
	 | C_SD.Command.Skip -> P.True
         | C_SD.Command.SymbolicHook _ -> P.True
         | C_SD.Command.Havoc lv -> 
	     failwith ("foci_post: Havoc unimplemented")
	 | C_SD.Command.HavocAll -> failwith "foci_post: HavocAll unimplemented"
	 | C_SD.Command.Phi (sym, flag) ->
	     failwith "foci_post: Phi unimplemented"
   in 
   match reg with
       Region.Atomic atomic_reg ->
	 let (new_location,new_stack) = location_post atomic_reg op  in
	 let post_reg = Region.Atomic { Region.location = new_location; 
					Region.stack = new_stack; 
					Region.data = Region.Abstract 
							{ Region.bdd = PredTable.bddOne;
                                                          Region.lelement = DataLattice.top ;
							  Region.preds = [] };
				      }
	 in
	   (post_reg,lvmap,op_constraint)
     | _ -> failwith "Unexpected region in foci_post!"





(****************************************************************************************************)
(**** focimodelchecker  refine trace [CONCRETE] *****)
(****************************************************************************************************)

(* focimodelchecker_refine_trace : 
   Region.t array -> 
     Operation.t array -> 
       P.predicate array -> 
         lvmap_t array -> Region.t list *)

(* where region is the present known region at the nodes. 
 * operation is the operation, and the pred list is the 
 * constraints of the Trace Formula -- precomputed while 
 * model checking, unlike in vanilla-blast *)

(* this function is very much like block_analyze_trace, described below. 
 * The difference is that we get the cons_l for free -- 
 * Q: is it a better idea to construct that on demand. Hmm. *)


let fociModelChecker_refine_trace 
  (reg_array: Region.t array) 
  (op_array: Operation.t array)
  (block_array: P.predicate array) 
  (lvmap_array_o : lvmap_t array option)
  =
  let len_reg_array = Array.length reg_array in
  let conj_block_array = Array.create len_reg_array P.True in  
  let _ = 
    let fill_cba i reg_i = 
      let i_conj = 
        match lvmap_array_o with
        Some (lvmap_array) -> 
          let pred_i = extract_pred reg_i in 
          (pred_lval_map_a (lvmap_array.(i)) 0 pred_i)
        | None -> P.True in
      Array.set conj_block_array i i_conj in
    Array.iteri fill_cba reg_array in
  let _ = M.msg_string M.Debug (Region.to_string (reg_array.(len_reg_array -1))) in
  if not (Options.getValueOfBool "blockfmc") then
    begin
      let return_reg_array =  Array.init len_reg_array (fun i -> reg_array.(i)) in
      let all_cons = (Array.to_list block_array) @ [conj_block_array.(len_reg_array - 1)] in
      let (ref_start_pos,inf_block) = 
        try
          let itp_list = (* Stats.time "extract foci itpN"*) (TheoremProver.extract_foci_interpolantN_noconv) all_cons in
          let _ = M.msg_string M.Debug ("back from itpN size = "^(string_of_int (List.length itp_list ))) in
          let min_i = ref 0 in
          let infb_l = ref [] in
          let rec folder i prev_pred itp_l = 
            match itp_l with 
              [] -> ()
            | h::t -> 
                let new_pred = P.unprime h in
                let p_id_l = List.map PredTable.addPred (P.getAtoms new_pred) in
                let i_reg = return_reg_array.(i) in
                let l_id = C_SD.location_coords (location_coords_of_region i_reg) in
                let _ = add_local_preds l_id p_id_l in
                let _ = if (new_pred <> P.True && !min_i = 0 ) then min_i := i in
                let _ = if (new_pred <> prev_pred) then infb_l := (i-1)::!infb_l in
                  Array.set return_reg_array i 
                (replace_data_region i_reg (Region.Concrete {Region.pred = new_pred;}));
                folder (i+1) new_pred t in
          let _ = folder 1 P.True itp_list in    
          (!min_i,!infb_l)
        with TheoremProver.FociSatException -> (-1,[])
          | TheoremProver.FociSatExceptionAsgn _ -> (-1,[])
          | e -> raise e in
      (ref_start_pos, Array.to_list return_reg_array,[inf_block])    
  end
  else (* use simplify to get useful blocks and then cut *)
    (* This is 1 greater then op_array *)
    let return_reg_array = Array.create len_reg_array Region.bot in
    let _ = M.msg_string M.Debug (P.toString (conj_block_array.(len_reg_array -1))) in
    (* now the conj_block_array is also filled *)
    let (ref_start_pos, interesting_block_sets, suppress_stuff) =
        get_useful_blocks op_array block_array conj_block_array in 
    if interesting_block_sets = [] then (-1, [], interesting_block_sets)
    else
      let (_,extract_block,_,_,_) = suppress_stuff in
      let fused_array = (Array.make (Array.length reg_array) P.True) in
      let _ = List.iter (interpolant_refine None (Some fused_array) extract_block conj_block_array reg_array) 
              interesting_block_sets in
      let _ = M.msg_string M.Debug "itp_array_l made" in
      (* will worry about normalizing and removing the "True"s in the fullness of time  *)
      let make_reg i i_reg = 
        let new_pred = P.unprime (fused_array.(i)) in
          Array.set return_reg_array i 
            (replace_data_region i_reg (Region.Concrete {Region.pred = new_pred;})) in 
      let _ = Array.iteri make_reg reg_array in
      (ref_start_pos, (Array.to_list return_reg_array ),interesting_block_sets)

        

(**********************************************************************************)
(******************** focimodelchecker [ABSTRACT] *********************************)
(**********************************************************************************)



 (************* cache the CONS constr. for each cfa-edge ***************)

let is_pure_skip cmd e_id (plist,plist') =
  if plist != plist' then false else
  match cmd.C_SD.Command.code with
        C_SD.Command.Pred e -> 
          let lvs = P.lvals_of_predicate (P.And [e; (pred_to_mccarthy e)]) in 
          not (List.exists PredTable.is_relevant_lval lvs)
      | _ -> not (List.exists (PredTable.is_pred_affected e_id) plist)
 
let (cons_table : (C_SD.edge_id_t,((int * int ) * P.predicate)) Hashtbl.t) 
        = Hashtbl.create 101

let predicate_update_constraints_cached e (postloc,ce_o) (in_plist,out_plist) = 
  let cmd = C_SD.get_command e in
  let e_id = (fst (C_SD.get_edge_coords e), C_SD.location_coords postloc) in
  let plist = Misc.sort_and_intersect (in_plist @ out_plist) in
  let this_ts = (List.length in_plist, List.length out_plist) in 
  let compute_and_update_cache () = 
    let pl2s x = Misc.strList (List.map P.toString x) in
    let (a,b,c) = 
      if O.getValueOfInt "focimax" = -1 then 
         predicate_update_constraints (cmd,e) postloc (in_plist,out_plist) 
      else block_predicate_update_constraints e (postloc,ce_o) (in_plist,out_plist) in
    let cp = 
      if (Options.getValueOfBool "focisymm" && (is_pure_skip cmd e_id (in_plist,out_plist))) 
      then P.True else P.conjoinL (a @ b @ c) in
    Hashtbl.replace cons_table e_id (this_ts,cp);
    M.msg_string M.Debug 
    (Printf.sprintf "pred_upd_cons: %s :\n %s \n %s \n %s" 
    (C_SD.edge_to_string e) (pl2s a) (pl2s b) (pl2s c));
    cp
  in
  try
    let (ts, res ) = Hashtbl.find cons_table e_id in
    if (O.getValueOfInt "focimax" <> -1 || (compare ts this_ts) = (-1)) (* cache is stale -- more predicates added *)
    then compute_and_update_cache ()
    else res
  with Not_found -> compute_and_update_cache ()
  
  
(* END CACHE *)

(*************************************************************)
(*********** Various Renaming Functions **********************)
(*************************************************************)

(**** timeframe  rename function p[i] 
replaces:
  bp_next_k --> bp_k_i+1,
  s --> s_i
************)
let interpreted_rename_suffix s e = 
  match e with 
    E.FunctionCall (E.Lval (E.Symbol fs),args) when E.is_interpreted fs ->

      let _ = M.msg_string M.Debug (Printf.sprintf "irs1: %s " (E.toString e)) in
      if not (E.is_interpreted_renamable fs) then (e,true)
      else 
        let lvs = List.flatten (List.map E.lvals_of_expression args) in
        let _ = M.msg_string M.Debug (Printf.sprintf "irs2: %s " (String.concat ";" (List.map E.lvalToString lvs))) in
        (match lvs with []-> (E.FunctionCall (E.Lval(E.Symbol(fs^s)),args),true) | _ -> (e,false))
  | E.Lval (E.Symbol x) when E.is_interpreted x -> 
      if not (E.is_interpreted_renamable x) then (e,true)
      else (E.Lval (E.Symbol (x^s)),true)
  | _ -> (e,true)

let rename_pred_timeframe i pred =
  let new_name suff suff' l = 
    match l with
      Expression.Symbol s ->
        let new_s =
          if E.is_interpreted s then s else 
            if (Misc.is_prefix bp_next_prefix s) then 
              let _s = Misc.chop_after_prefix bp_next_prefix s in
              (bp_prefix^_s^suff')
            else (s^suff)
      in Expression.Symbol new_s
    | _ -> failwith ("Bad lval rename_pred_timeframe: "^(Expression.lvalToString l))    
  in
  let is = "_"^(string_of_int i) in
  let p = P.symbol_alpha_convert (new_name is ("_"^(string_of_int (i+1)))) pred in
  let rv = P.atom_transform (E.deep_transform_top_cond (interpreted_rename_suffix is)) p in
  M.msg_string M.Debug (Printf.sprintf "rename_pred_timeframe: in = %s , out = %s " (P.toString p) (P.toString rv));
  rv


let remove_pred_timeframe  pred = 
  let new_name l = 
    match l with
      Expression.Symbol s -> 
        let pieces = Misc.chop s "_" in
        let _ = assert (pieces <> []) in
        let s' = if (pieces <> []) then String.concat "_" (Misc.chop_last pieces) else s in
        Expression.Symbol s'
    | _ -> failwith ("Bad lval remove_pred_timeframe: "^(Expression.lvalToString l))    
  in
  P.symbol_alpha_convert new_name  pred

(**** convert a formula using atomic predicates into a boolean formula *)
(* p should use atoms corresponding to existing predicates.
   This function returns pred[atom_i -> bp_prefix_i] *)
  
let lookup_atom_index p = 
  try (PredTable.getPredIndex p,true) 
  with Not_found -> (PredTable.getPredIndex (P.negate p),false)

(* requires unprimed, canonical pred, with atoms known to PredTable *)
let convert_pred_to_boolean pred =
  let tf p = 
    match p with P.Atom e -> 
        let (i,b) = try lookup_atom_index p 
                    with Not_found -> failwith ("unknown atom in conv_pred_to_bool:"^P.toString p) in
        let p' = P.Atom (Expression.Lval (Expression.Symbol (bp_prefix^(string_of_int i)))) in
        if b then p' else (P.negate p')
    | _ -> p in
  P.deep_transform tf pred



let convert_boolpred_to_pred j bp =
  let tf p = 
    match p with
    P.Atom e ->
      begin
         match e with 
           Expression.Lval (Expression.Symbol s) -> 
           let (time_s::idx_s::_)  = 
             try
                if (Misc.is_prefix bp_prefix s) then (* sanity check *) 
                   let pieces = (Misc.chop s "_") in
                   List.rev pieces 
                else failwith (Printf.sprintf "Bad symbol in convert_boolpred_to_bdd: %s : %s" s (P.toString bp))
             with _ -> failwith ("Bad arg to convert_boolpred_to_bdd - symbol: "^s)
           in
           let time = int_of_string time_s in
           let idx = int_of_string idx_s in
           if (time = j) then PredTable.getPred idx
           else if (time = j+1) then (P.Next (PredTable.getPred idx))
           else failwith (Printf.sprintf "Bad atom in convert_boolpred_to_pred: %d %s" j (P.toString bp))
          |  _ -> failwith ("Bad arg to convert_boolpred_to_pred -atom: "^(Expression.toString e))
        end
    | _ -> p 
  in
  P.deep_transform tf bp 

(* takes a predicate made of bp_i variables (ONLY) turns it into a bdd
   replace bp_i_j with bdd_i, replace bp_i_{j+1} with bdd_i' *)

let convert_boolpred_to_bdd j pred = 
   let tf p = 
     match p with
        P.Atom e ->
        begin
         match e with 
           Expression.Lval (Expression.Symbol s) -> 
           let (time_s::idx_s::_)  = 
             try
                if (Misc.is_prefix bp_prefix s) then (* sanity check *) 
                   let pieces = (Misc.chop s "_") in
                   List.rev pieces 
                else failwith (Printf.sprintf "Bad symbol in convert_boolpred_to_bdd: %s : %s" s (P.toString pred))
             with _ -> failwith ("Bad arg to convert_boolpred_to_bdd - symbol: "^s)
           in
           let time = int_of_string time_s in
           let idx = int_of_string idx_s in
           if (time = j) then PredTable.getPredIndexBdd idx
           else if (time = j+1) then PredTable.getPredIndexNextBdd idx
           else failwith (Printf.sprintf "Bad timestamp in convert_boolpred_to_bdd: %d %s" j (P.toString pred))
          |  _ -> failwith ("Bad arg to convert_boolpred_to_bdd -atom: "^(Expression.toString e))
        end
    | _ -> failwith ("[convert_boolpred_to_bdd] bad call to tf: "^(P.toString p)) 
  in
  PredTable.convertPredToBdd_map tf pred
 

(* as of now, this turns the thing into DNF so be careful *)  
let convert_bdd_to_boolpred bdd =
   let idx_to_string i = 
     if isCurrentIndex i then (bp_prefix^(string_of_int i)) 
     else (bp_next_prefix^(string_of_int i)) 
   in
   (* PredTable.bdd_to_pred_map 
   (fun i -> PredTable.string_to_atom (idx_to_string i))
   bdd *)
   ac_convertBddToPred 
   (fun i -> P.Atom (Expression.Lval (Expression.Symbol (idx_to_string i)))) 
   bdd 

let trel_bdd_to_string trel_bdd = 
  let bp = convert_bdd_to_boolpred trel_bdd in
  let bp0 = rename_pred_timeframe 0 bp in
  let cp = convert_boolpred_to_pred 0 bp0 in 
  P.toString cp
  


(**************************************) 
 (** This is there in some form in the above, but here's a clearer explanation.
 * Trace: R_0 -- op_0 -> R_1 ... -> R_i -- op_i --> R_i+1 ... -> R_n 
 * Define:
 *   Cons[i] = CHAIN[i] /\ WPC[i] /\ R_i[i]  where:
 *   CHAIN   =   /\_p (b <=> p)
 *   WPC[i]  = /\_p ( (WP.p.op_i)[i] <=> b[i+1])
 * We want: 
 *   1. R'_i = ITP.(A_i,B_i)
 *   where A_i = /\_(0<=j<=i-1) Cons[j] ,
 *         B_i = /\_(i<=k<=n) Cons[k] 
 *   2. T'_i = ITP.(Cons[i], _rest_)
 *)

(* A few global structures *)

let foci_ctr = ref 0

(* Gathering the foci theory clauses -- tautologies that may make reasoning fully boolean *)
let foci_theory_clauses_ref = ref []

(* for now, just accumulate them *)
let accumulate_theory_clauses (tc_l_l :P.predicate list list) =
   let _ = M.msg_string M.Debug "In accumulate_theory_clauses" in
   let proc_tc tc =
     let tc_notime = remove_pred_timeframe tc in
     foci_theory_clauses_ref := tc_notime :: (!foci_theory_clauses_ref);
     let s = List.length !foci_theory_clauses_ref in
     M.msg_string M.Debug (Printf.sprintf "proc_theory_clauses: %d : %s" s (P.toString tc_notime))
   in
   List.iter proc_tc (List.flatten tc_l_l)
   
let iter_theory_clauses f =
  M.msg_string M.Normal 
  ("Num theory clauses: "^(string_of_int (List.length !foci_theory_clauses_ref)));
  List.iter f !foci_theory_clauses_ref
  (* ignore (List.map (fun p -> M.msg_string M.Normal
  (P.toString p)) !foci_theory_clauses) *)

let print_counterexample op_array rev_map_array plist =
 let _ = M.msg_string M.Normal ("SAT preds: "^(Misc.strList plist)) in
 let ptable = Hashtbl.create 101 in
 let add_to_table s =
   try
   if (Misc.is_prefix bp_prefix s) then 
     let pieces = Misc.chop s "_" in
     let (ps,ts) = Misc.last_two pieces in
     let p = int_of_string ps in
     let t = try rev_map_array.(int_of_string ts) with _ -> failwith "check print_counterexample" in
     Misc.hashtbl_update ptable t p;
   with _ -> ()
 in
 let _ = List.iter add_to_table plist in
 let cubes_l_ref = ref [] in
 let print_i i op_i = 
   let (l1,l2,rv) = op_to_string op_i in
   let pred_s =
     let pi_l = try Hashtbl.find ptable i with Not_found -> [] in
     let p_l = List.map PredTable.getPred pi_l in
     Misc.strList (List.map P.toString p_l)
   in
   M.msg_string M.Normal (Printf.sprintf "Time: %d : %s " i pred_s);
   M.msg_string M.Normal (Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2);
   cubes_l_ref := pred_s :: !cubes_l_ref
 in
 M.msg_string M.Normal "***************Abstract Counterexample**********";
 Array.iteri print_i op_array;
 let rv = List.rev ("True"::!cubes_l_ref) in
 htmlize_trace (Array.to_list op_array) (List.tl rv) [] "Abstract Counterexample";
 rv
 
   
let check_tautology bp = 
  let gather_timeframes p = 
    let tfms = ref [] in
    let _tf p = 
      match p with
      P.Atom e ->
       begin
        match e with 
        Expression.Lval (Expression.Symbol s) -> 
          (try 
            let ts = int_of_string (Misc.list_last (Misc.chop s "_")) in 
            tfms := ts::!tfms;
            p
          with _ -> (failwith ("Bad atom1 in gather_timeframes  :"^(P.toString p))))
        | _ ->  (failwith ("Bad atom2 in gather_timeframes  :"^(P.toString p)))
       end 
     | _ -> p
   in
   P.deep_transform _tf bp;
   !tfms
 in
 if (O.getValueOfBool "mccarthy") then (false,PredTable.bddOne) else
 match gather_timeframes bp with
    [] -> (false,PredTable.bddOne)
   |[i] -> 
       let p = P.Not (convert_boolpred_to_pred i bp) in
       let rv = TheoremProver.queryExp (P.Implies (p, P.False)) in
       let rvbdd = if rv then convert_boolpred_to_bdd i bp else PredTable.bddOne in
       (rv,rvbdd)
   | _ -> (false,PredTable.bddOne)
     
let rec process_interpolants_taut i p_list =
    match p_list with
     [] -> ()
   | h::t -> 
     let cp = convert_boolpred_to_pred i h in
    M.msg_string M.Normal (P.toString cp);
    let (is_taut,tbdd) = check_tautology h in
    let pflag = if is_taut then M.Normal else M.Debug in
    M.msg_string pflag (Printf.sprintf "Tautology?: %s : %b" (P.toString cp) is_taut);
    if (is_taut) then update_global_inv_bdd tbdd;
    process_interpolants_taut (i+1) t


(* A few "quick" ways to strengthen the trel using foci *)
  
let strengthen_trel i (l_id,l_id') (bdd,bdd') (ibdd,ibdd') (itp,itp') op_cons_timed = 
  (* 1. Copy the values of unchanged predicates *)
  let copy_trel_bdd = 
    let pred_i_list = PredTable.bdd_support ibdd' in
    let unch_p_i_list = 
      List.filter (fun i -> not (PredTable.is_pred_affected (l_id,l_id') i))  pred_i_list 
    in
    M.msg_string M.Debug ("Unchanged Preds: "^(Misc.string_of_int_list unch_p_i_list));
    PredTable.make_eq_bdd unch_p_i_list
  in
  let expensive_bdd () = 
  (* 2. A strengthening function -- input: upper bound on postred output*)
  let curr_pred = 
    let joined_bdd = CaddieBdd.bddAnd bdd ibdd in
    let jbd_cartabs_l = PredTable.bdd_cartesian_decomposition joined_bdd in
    let bpred_l = 
      List.map 
        (fun b -> rename_pred_timeframe i (convert_bdd_to_boolpred b)) 
        jbd_cartabs_l 
    in
    P.And bpred_l
  in
  let make_cons_bdd next_pred_bdd = 
    let make_new_curr next_pred_timed = 
     let itp_l = 
      try TheoremProver.extract_foci_interpolantN 
         [curr_pred; op_cons_timed; (P.negate next_pred_timed)]
      with _ -> failwith "Foci says sat or fails -- issue in strengthen"
     in
     List.hd itp_l
    in
    let np_timed = rename_pred_timeframe (i+1) (convert_bdd_to_boolpred next_pred_bdd) in
    let timed_trel_cons = 
      try
        let cp_new = make_new_curr np_timed in
        P.Implies(cp_new, np_timed) 
      with _ -> (M.msg_string M.Normal "Strengthen fails!"; P.True)
    in
    let _ = M.msg_string M.Debug ("st. constraint:"^(P.toString timed_trel_cons)) in
    convert_boolpred_to_bdd i timed_trel_cons
  in
 (* 3. (De)compose the post region into conjuncts (collection of upper bounds),
        and then strengthen the trel for each conjunct *)
  let _ = M.msg_string M.Normal "Now decomposing" in
  let _ = 
    let bb = CaddieBdd.bddEqual bdd bdd' in
    let ibb = CaddieBdd.bddEqual ibdd ibdd' in
    M.msg_string M.Debug (Printf.sprintf "[Equal] bdd: %b ibdd: %b" bb ibb)
  in
  let (p'_list) = 
    let c' = CaddieBdd.bddAnd bdd' ibdd' in
    let _ = M.msg_string M.Debug 
    (Printf.sprintf "[C'] \n  bdd': %s \n ibdd': %s"
    (PredTable.bdd_to_string bdd')
    (PredTable.bdd_to_string ibdd'))
    
  in
    let rv = PredTable.bdd_cartesian_decomposition c' in
    let _ = M.msg_string M.Debug 
    (Printf.sprintf "[CDEC] \n\n In:\n %s \n\n  Out: \n %s" 
    (PredTable.bdd_to_string c') 
    (Misc.string_list_cat "\n" (List.map PredTable.bdd_to_string rv))) in
    rv
  in
  
  let _ = M.msg_string M.Normal "Now finding copy" in
  (* Q: We could do it for just the predicates in p'_list that are in the support of ibdd'... but *)    
  (*let copied_indices =  
        let b_supp = PredTable.bdd_support ibdd in
        let b'_supp = PredTable.bdd_support ibdd' in 
        Misc.sort_and_intersect b_supp b'_supp
  in      
  let _ = M.msg_string M.Normal "Now calling bdd_exist_quantify" in
  let ep' = PredTable.bdd_exist_quantify p' copied_indices in*)
  Misc.accumulate  make_cons_bdd CaddieBdd.bddAnd bddOne (p'_list)
      
  in (* done making expensive bdd *)
  
  if (Options.getValueOfInt "trelc" = 2) then 
    let rv = expensive_bdd () in
    CaddieBdd.bddAnd copy_trel_bdd rv
  else copy_trel_bdd    


  

let process_inconsistent_ops op_array asm_mask_array op_id_l =
  let proc () i = 
    let op_i = op_array.(i) in
    let (l1,l2,rv) = op_to_string op_i in
    M.msg_string M.Normal (Printf.sprintf "[INF%d] %d :: %s :: %d \n\n" !foci_ctr l1 rv l2);
    if (C_SD.isPredicate op_i) then (Array.set asm_mask_array i true);
  in
  List.fold_left proc () op_id_l;
  M.msg_string M.Normal (Printf.sprintf "[INF%d] \n\n\n" !foci_ctr)


  
exception TrelUnsatExn of int

let fill_ctrex_cubes cc_array reg_array = 
  M.msg_string M.Debug "fill ctrex cubes";
  let n = Array.length reg_array in
  Array.set cc_array (n-1) (PredTable.extract_full_cube (extract_abs_data_reg reg_array.(n-1)).Region.bdd); 
  let idx = ref (n-2) in
  while (!idx >= 0) do
    begin
      let next_cube = cc_array.(!idx + 1) in
      if (PredTable.bdd_cubes next_cube > 1) then 
        (M.msg_string M.Error "Warning >1 cube in next_cube!";
        M.msg_string M.Normal "Warning >1 cube in next_cube!");
      let trel_bdd = 
        let this_loc = C_SD.location_coords (location_coords_of_region (reg_array.(!idx))) in
        let next_loc = C_SD.location_coords (location_coords_of_region (reg_array.(!idx + 1))) in
        PredTable.get_transrel_bdd (this_loc,next_loc) 
      in
      let idxbdd = (extract_abs_data_reg reg_array.(!idx)).Region.bdd in
      let this_bdd_int = 
        CaddieBdd.bddAnd (PredTable.bdd_preimage next_cube trel_bdd) idxbdd
      in
       if (CaddieBdd.bddEqual idxbdd PredTable.bddZero) then
        M.msg_string M.Normal ("Trel-unsat-here1! "^(string_of_int !idx));
      if (CaddieBdd.bddEqual next_cube PredTable.bddZero) then
        M.msg_string M.Normal ("Trel-unsat-here! "^(string_of_int !idx));
      if (CaddieBdd.bddEqual this_bdd_int PredTable.bddZero) then
        (M.msg_string M.Normal ("Trel-unsat! "^(string_of_int !idx));
        raise (TrelUnsatExn (!idx)))
      else Array.set cc_array !idx (PredTable.extract_full_cube this_bdd_int);
      idx := !idx - 1
    end;
  done;
  () 

let printop i a = 
 let (_,_,op_s) = op_to_string a in
 M.msg_string M.Debug (Printf.sprintf "%d : %s" i op_s)

let print_tpso tps_o =
  M.msg_string M.Debug "Printing tps_o: \n";
  match tps_o with None -> M.msg_string M.Debug "tps_o: empty"
  | Some a -> Array.iteri (fun i l -> M.msg_string M.Debug (Printf.sprintf "%d : %s \n" i (Misc.string_of_int_list l))) a 
  
let build_cons aps clock_ref cpc reg_array cefn conj_block_array copy_conj_block_array rev_map_array tps_o i op_i = 
  M.msg_string M.Debug "in build_cons";
  let thisloc = location_coords_of_region (reg_array.(i)) in
  let postloc = location_coords_of_region (reg_array.(i+1)) in
  let rel_preds_in = match tps_o with Some a -> (print_string "some tps_o" ; Misc.sort_and_compact a.(i)) 
   | None -> 
      if (Options.getValueOfInt "craig" = 2) then PredTable.lookup_location_predicates (C_SD.location_coords thisloc) 
      else aps in
  let rel_preds_out = match tps_o with Some a -> (print_string "some tps_o"; Misc.sort_and_compact a.(i+1)) 
   | None ->
      if (Options.getValueOfInt "craig" = 2) then PredTable.lookup_location_predicates (C_SD.location_coords postloc) 
      else aps in
  let _ = 
    M.msg_string M.Debug (Printf.sprintf "build_cons: op: %d : %s : in_p : %s : out_p : %s \n"
    i (Misc.thd3 (op_to_string op_i)) (Misc.string_of_int_list rel_preds_in)
    (Misc.string_of_int_list rel_preds_out))
  in
  let ce_o = cefn i postloc in
  let chain_wpc = predicate_update_constraints_cached op_i (postloc,ce_o) (rel_preds_in,rel_preds_out) in
  if (chain_wpc <> P.True) then (* Non-trivial edge, generate constr. *)
  begin 
    let r = 
      if (Options.getValueOfBool "regcons") then convert_pred_to_boolean (extract_pred (reg_array.(i))) 
      else P.True in
    let cp = P.conjoinL [chain_wpc;r] in
    let cons_i = rename_pred_timeframe (!clock_ref) cp in
    let copy_i = rename_pred_timeframe (!clock_ref) cpc in
       Array.set conj_block_array (!clock_ref) cons_i;
       Array.set copy_conj_block_array (!clock_ref) copy_i;
       Array.set rev_map_array !clock_ref i;
       clock_ref := !clock_ref + 1
  end
  

let build_fmc_constraint_arrays reg_array op_array tps_o =
 let len_reg_array = Array.length reg_array in 
 let _ = M.msg_string M.Debug ("build_fmc_cons: len_reg_array = "^(string_of_int len_reg_array)) in
 assert (len_reg_array = Array.length op_array + 1);
 Array.iteri printop op_array;
 let conj_block_array = Array.create len_reg_array P.True in
 let copy_conj_block_array = Array.create len_reg_array P.True in
 let ctrx_cube_bdd_array = Array.create len_reg_array PredTable.bddOne in
 let ctrx_cube_array = Array.create len_reg_array P.True in
 let rel_pred_i_list = PredTable.get_all_pred_indexes () in
 let copy_pred_constr = copy_update_constraints rel_pred_i_list in
 let rev_map_array = Array.create len_reg_array (-1) in  
 let clock_ref = ref 0 in
 (* {{{ let tps_o = 
   if O.getValueOfBool "lazysymm" then 
     (M.msg_string M.Debug "build_fmc_cons: calling ac_block_analyze_trace";
     let (_,_,_,x,_) = ac_block_analyze_trace (Array.to_list reg_array) (Array.to_list op_array) in 
     ignore(print_tpso x); x)
  else None in }}}  *)
 let cefn = get_callEdge_o op_array in
 for i = 0 to (Array.length op_array - 1) do
   (build_cons rel_pred_i_list clock_ref copy_pred_constr) 
    reg_array cefn conj_block_array copy_conj_block_array rev_map_array tps_o i (op_array.(i))
 done;  
 let n = !clock_ref in
 let r_n =  rename_pred_timeframe n (convert_pred_to_boolean (extract_pred (reg_array.(len_reg_array -1)))) in   
 Array.set conj_block_array n r_n;
 let cxcube_s_l = List.map (fun c -> P.toString (convertBddToPred c)) (Array.to_list ctrx_cube_bdd_array) in
 M.msg_string M.Debug "Non-skip constraints";
 for i = 0 to (!clock_ref - 1) do
   let ac_i = rev_map_array.(i) in 
   printop ac_i op_array.(ac_i)
 done;
 (conj_block_array,copy_conj_block_array,rev_map_array,!clock_ref,ctrx_cube_array,cxcube_s_l)

 
let get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops =
  let rec gather idx l =
    if idx > num_ops then (List.rev l)
    else 
      let h = if asm_mask_array.(idx) then copy_conj_block_array.(idx) else conj_block_array.(idx) in
      gather (idx+1) (h::l) in
  gather 0 []

let refine_reach_states reg_array op_array (* conj_block_array copy_conj_block_array *) asm_mask_array itp_array return_reg_array (* num_ops *) =
  let len_reg_array = Array.length reg_array in 
  let _ = M.msg_string M.Debug (Printf.sprintf "in refine_reach_states %d" len_reg_array) in
  (* let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops in*)
  let preproc i itp = 
    let itp' = P.eq_canonicize (P.unprime itp) in
    ignore(List.map PredTable.addPred (P.getAtoms itp'));
    rename_pred_timeframe (i+1) (convert_pred_to_boolean itp') in
  let inc_op_list = ref [] in
  let min_i = ref (-1) in
  let rec folder i itp_l = 
    match itp_l with [] -> () 
    | h::t ->
        M.msg_string M.Debug ("Access array at:"^(string_of_int i));
        let _ = Array.set itp_array i h in
        let new_i_bdd  = convert_boolpred_to_bdd i h in
        let _ = if (not (CaddieBdd.bddEqual new_i_bdd PredTable.bddOne) && (!min_i = -1)) 
                then  min_i := i in
        let new_i_reg = insert_bdd_reg (return_reg_array.(i)) new_i_bdd in
        Array.set return_reg_array i new_i_reg;
        let last_bdd = (extract_abs_data_reg (return_reg_array.(i-1))).Region.bdd in
        let _ = if not (CaddieBdd.bddEqual last_bdd new_i_bdd) then inc_op_list := (i-1)::!inc_op_list; in 
        let l_id= C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
        let this_preds = PredTable.bdd_support new_i_bdd in
        let _ = PredTable.add_local_preds l_id this_preds in
              (* {{{ let restricted_itp_bdd = 
                let reach_cube = PredTable.bdd_cube_cover ( (extract_abs_data_reg (reg_array.(i))).Region.bdd) in
                CaddieBdd.bddAnd reach_cube new_i_bdd
              in
              M.msg_string M.Normal 
              (Printf.sprintf "itp-cubes: (%d,%d) =  %d restricted %d" (fst l_id)
              (snd l_id) (PredTable.bdd_cubes new_i_bdd)
              (PredTable.bdd_cubes restricted_itp_bdd)); }}} *)
        folder (i+1) t in
  let (rsp,_,_,_,acc_itp_o) = ac_block_analyze_trace (Array.to_list reg_array) (Array.to_list op_array) in
  if rsp < 0 
  then (-1,[]) (* raise TheoremProver.FociSatException*)
  else
    (match acc_itp_o with  None -> assert false
      | Some a -> (folder 1 (Misc.mapi preproc (List.tl (Array.to_list a))));
      (* PredTable.print_pred_vector ();PredTable.dump_abs ();*)
      (rsp,!inc_op_list))  
     

let refine_trel_states reg_array conj_block_array copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops =
  failwith "refine_trel_states : obsolete";
  let len_reg_array = Array.length reg_array in 
  let cheap_extract (src_id,dst_id) i cons_i = 
    let itp_b = (extract_abs_data_reg (return_reg_array.(i))).Region.bdd in
    let itp_b' = (extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd in
    let b = (extract_abs_data_reg (reg_array.(i))).Region.bdd in
    let b' = (extract_abs_data_reg (reg_array.(i+1))).Region.bdd in
    let itp = (itp_array.(i)) in
    let itp' = (itp_array.(i+1)) in
    M.msg_string M.Debug "Calling strengthen trel";
    strengthen_trel i (src_id,dst_id) (b,b') (itp_b,itp_b') (itp,itp') cons_i in
  let rec upd_trel i cons_past cons_future =
    match cons_future with
    [] -> failwith "empty list in refine trel"
    | [h] -> () (* the target region *)
    | cons_i::cons_future' ->
        M.msg_string M.Debug (Printf.sprintf "Refining edge %d of %d" i len_reg_array);
        let rrbdd = (extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd in
        if (not (CaddieBdd.bddEqual rrbdd PredTable.bddOne)) then 
          (* this op participated in proof of unsat *)
          begin
            let src_id = C_SD.location_coords (location_coords_of_region (return_reg_array.(i))) in
            let dst_id = C_SD.location_coords (location_coords_of_region (return_reg_array.(i+1))) in
            let itp_b = 
              let b_cheap = cheap_extract (src_id,dst_id) i cons_i in
              let b_exp =     
                let _ = M.msg_string M.Debug "trel:foci" in
                let itp =
                  let a = cons_i in
                  let b = 
                    if (Options.getValueOfInt "trelc" = 0) then (cons_past @ cons_future')
                    else ([itp_array.(i) ; P.negate itp_array.(i+1)]) in
                  try List.hd (TheoremProver.extract_foci_interpolantN [a;(P.And b)]) 
                  with e -> 
                    M.msg_string M.Error ("Foci says:"^(Printexc.to_string e));
                    M.msg_string M.Error "WARNING! : BAD ITP!";
                    (P.Not (P.And b)) in
                convert_boolpred_to_bdd i itp in
              CaddieBdd.bddAnd b_cheap b_exp in
            PredTable.update_transrel_bdd (src_id,dst_id) itp_b;
            (* check if this is a trivial interpolant *)
            let _ = 
              let triv_trel_bdd = 
              let b_i = (extract_abs_data_reg (return_reg_array.(i))).Region.bdd in
              let b_i1' = PredTable.make_primed_bdd ((extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd) in
              CaddieBdd.bddImp b_i b_i1' in
            if (CaddieBdd.bddEqual (itp_b) triv_trel_bdd) then M.msg_string M.Normal "TREL-TRIV"
            else M.msg_string M.Normal "TREL-NON-TRIV" in
            let b_i = (extract_abs_data_reg (return_reg_array.(i))).Region.bdd in
              let b_i1 = (extract_abs_data_reg (return_reg_array.(i+1))).Region.bdd in
              let b_i1' = CaddieBdd.replace b_i1 (!PredTable.next_var_map) in
              let b_i' = CaddieBdd.bddAndAbstract b_i itp_b (!PredTable.curr_var_cube)  in
              let leqcheck = CaddieBdd.bddEqual (CaddieBdd.bddImp b_i' b_i1') PredTable.bddOne in
              if not (leqcheck) then 
                begin
                  M.msg_string M.Error ("bdd-itp failure! at i = "^(string_of_int i));
                  M.msg_string M.Debug ("bdd-itp failure! at i = "^(string_of_int i));
                  M.msg_string M.Debug (Printf.sprintf "b_i = %s" (PredTable.bdd_to_string b_i));
                  M.msg_string M.Debug (Printf.sprintf "itp_b = %s" (PredTable.bdd_to_string itp_b));
                  M.msg_string M.Debug (Printf.sprintf "b_i' = %s" (PredTable.bdd_to_string b_i'));
                  M.msg_string M.Debug (Printf.sprintf "b_i1' = %s" (PredTable.bdd_to_string b_i1'));
                  M.msg_string M.Debug "Done bdd print"
                end;
              () 
              end;
            upd_trel (i+1) (cons_i::cons_past) cons_future'
  in
  M.msg_string M.Debug "Now refining transition relation";
  let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops
  in
  upd_trel 0 [] all_cons
 (* end refine_trel_states *)

let check_symm_itp trel_cons itp =               
  if (TheoremProver.interpolant_sat (trel_cons, P.Not itp) = TheoremProver.Sat) 
  then 
    begin
      M.msg_string M.Error ("Bogus itp: "^(P.toString itp));
      M.msg_string M.Error ("For: ");
      M.msg_string M.Error (P.toString trel_cons);
      failwith "Invalid symm interpolant!"
    end

let print_update_trel src_id dst_id p = 
  M.msg_string M.Debug 
   (Printf.sprintf "upd_trel: (%d,%d) ---> (%d,%d) : %s" 
     (fst src_id) (snd src_id) (fst dst_id) (snd dst_id) (P.toString p))

let make_eq_pred plist = 
  P.And (List.map (fun i -> let p = PredTable.getPred i in P.Iff (p,P.Next p)) plist)

let update_skip_edges reg_array start_idx stop_idx plist =
  let cp_bdd = PredTable.make_eq_bdd plist in
  let cp_pred = make_eq_pred plist in
  for i = start_idx to stop_idx do
    let src_id = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
    let dst_id = C_SD.location_coords (location_coords_of_region (reg_array.(i+1))) in
    print_update_trel src_id dst_id cp_pred;              
    PredTable.update_transrel_bdd (src_id,dst_id) cp_bdd
  done

let strengthen_trel_symm reg_array rev_map_array cons_l itp_l =
  assert (List.length cons_l = List.length itp_l); 
  let min_i = ref (-1) in
  let inc_op_l = ref [] in
  let rec folder i_clock cons_l itp_l = 
    match (cons_l, itp_l) with (_,[]) -> () | (_,[h]) -> () 
    | (ch::ct, h::t) ->
        (let i = rev_map_array.(i_clock) in
          (if (h <> P.True) then
            let _ = if (Options.getValueOfBool "checktree") then check_symm_itp ch h in
            if !min_i = -1 then min_i := i;
            let _ = inc_op_l := i::!inc_op_l in
            let src_id = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
            let dst_id = C_SD.location_coords (location_coords_of_region (reg_array.(i+1))) in
            let trel_i_bdd = convert_boolpred_to_bdd i_clock h in
            let trel_i_pred = convert_boolpred_to_pred i_clock h in
            print_update_trel src_id dst_id trel_i_pred;              
            PredTable.update_transrel_bdd (src_id,dst_id) trel_i_bdd;
            let (inpreds,outpreds) = 
              let _preds = (PredTable.bdd_support trel_i_bdd) in
                (List.filter PredTable.isCurrentIndex _preds, 
                Misc.map_partial PredTable.getCIndexFromNIndex _preds) in
            ignore(PredTable.add_local_preds dst_id outpreds);
            (if (i_clock > 0) then update_skip_edges reg_array (rev_map_array.(i_clock-1)+1) (i-1) inpreds););
          folder (i_clock + 1) ct t) in 
  folder 0 cons_l itp_l ;
  ((if (!min_i = -1) then (-1) else (!min_i + 1)), !inc_op_l)

let get_lazysymm_itps cons_l rma itpa = 
  let _ = M.msg_string M.Normal "in get_lazysymm_itps" in
  let ren i p =
    let p = P.eq_canonicize (P.unprime p) in
    let pids = List.map PredTable.addPred (P.getAtoms p) in
    (pids,rename_pred_timeframe i (convert_pred_to_boolean p)) in
  let rec _f i_clock cs rs = 
    match cs with 
      [] -> List.rev rs 
    | [_] -> List.rev (P.True::rs)
    | (c::cs') ->
        let i = rma.(i_clock) in 
        let (ids_in,ri) = ren i_clock itpa.(i) in
        let (ids_out,ri') = ren (i_clock + 1) itpa.(i+1)  in
        let (rv::_,_) = (* TBD:OPTIMIZE *)
          Stats.time "get symm slice" TheoremProver.extract_foci_interpolantN_symm ([c;P.And [ri;P.Not ri']]) in
        _f (i_clock + 1) cs' (rv::rs) in
  try _f 0 cons_l [] with e -> failwith ("error in get_lazysymm_itps :"^(Printexc.to_string e))

(* may throw SAT exn *)
let refine_trel_symm reg_array conj_block_array copy_conj_block_array asm_mask_array rev_map_array num_ops acc_itp_o = 
  let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops in
  let all_itp_l = 
    match (O.getValueOfBool "slicesymm", acc_itp_o)  with 
      (true, Some a) -> Stats.time "get lazysymm itp" (get_lazysymm_itps all_cons rev_map_array) a 
    | _ -> fst (Stats.time "get symm itp" TheoremProver.extract_foci_interpolantN_symm all_cons) in 
  Stats.time "strengthen_trel_symm" (strengthen_trel_symm reg_array rev_map_array all_cons) all_itp_l

(* {{{

 let refine_trel_symm reg_array conj_block_array copy_conj_block_array asm_mask_array rev_map_array num_ops acc_itp_o = 
  let min_i = ref (-1) in
  let inc_op_l = ref [] in
  (* {{{ let curr_bdd1_ref = ref (PredTable.bddOne) in
  let curr_bdd2_ref = ref (PredTable.bddOne) in
let curr_trel_ref = ref [] in }}} *)
  let rec folder i_clock cons_l itp_l = 
    let _ = assert (List.length cons_l = List.length itp_l) in
    match (cons_l, itp_l) with (_,[]) -> () | (_,[h]) -> () 
    | (ch::ct, h::t) ->
        begin
          let i = rev_map_array.(i_clock) in
          if (h <> P.True) then
            begin
              let _ = if (Options.getValueOfBool "checktree") then check_symm_itp ch h in
              if !min_i = -1 then min_i := i;
              let _ = inc_op_l := i::!inc_op_l in
              let src_id = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
              let dst_id = C_SD.location_coords (location_coords_of_region (reg_array.(i+1))) in
              let trel_i_bdd = convert_boolpred_to_bdd i_clock h in
              let trel_i_pred = convert_boolpred_to_pred i_clock h in
              print_update_trel src_id dst_id trel_i_pred;              
              PredTable.update_transrel_bdd (src_id,dst_id) trel_i_bdd;
              let (inpreds,outpreds) = 
                let _preds = (PredTable.bdd_support trel_i_bdd) in
                  (List.filter PredTable.isCurrentIndex _preds, 
                  Misc.map_partial PredTable.getCIndexFromNIndex _preds) in
              ignore(PredTable.add_local_preds dst_id outpreds);
              (if (i_clock > 0) then update_skip_edges reg_array (rev_map_array.(i_clock-1)+1) (i-1) inpreds);
              if (Options.getValueOfBool "checktree" && not (Options.getValueOfBool "dasdill")) then
                begin
                  let trel_bdd = PredTable.get_transrel_bdd (src_id,dst_id) in
                  curr_bdd1_ref := PredTable.bdd_postimage !curr_bdd1_ref trel_bdd;
                  curr_bdd2_ref := PredTable.bdd_postimage !curr_bdd2_ref trel_i_bdd;
                  (* bdd1 => bdd2 *)
                  if not (CaddieBdd.bddEqual (CaddieBdd.bddImp !curr_bdd1_ref !curr_bdd2_ref) PredTable.bddOne)
                  then (failwith "Bdd-image interpolant issue (IMP)");
                  if (Options.getValueOfBool "bddprint") then
                    begin
                      M.msg_string M.Debug "Using TREL:";
                      M.msg_string M.Debug (trel_bdd_to_string trel_bdd);
                      M.msg_string M.Debug (Printf.sprintf "Next abs. state at %d#%d" (fst dst_id) (snd dst_id));
                      M.msg_string M.Debug (P.toString (convertBddToPred !curr_bdd1_ref)) 
                end  
            end  
            end;
          folder (i_clock + 1) ct t 
        end in
  let all_cons = get_cons asm_mask_array copy_conj_block_array conj_block_array num_ops in
  let (all_itp_l,_ (* tc_l_l *) ) = 
     if (Options.getValueOfBool "dasdill") then
      let get_res () = 
      let cube_cons = Misc.array_to_list_n ctrx_cube_array (num_ops + 1) in
      let trel_cons = Misc.chop_last all_cons in
      let dasdill_itp_l = TheoremProver.extract_foci_interpolantN_dasdill trel_cons cube_cons in
      if (Options.getValueOfBool "checktree") then List.iter2 check_symm_itp trel_cons dasdill_itp_l;
      (dasdill_itp_l @ [P.True],[])
      in 
      Stats.time "get dd itp" get_res () else  
    (Stats.time "get symm itp" TheoremProver.extract_foci_interpolantN_symm all_cons) in (* may throw SAT exn *)
  let _ = 
    if (Options.getValueOfBool "checktree" && not (Options.getValueOfBool "dasdill") &&
    (TheoremProver.interpolant_sat (P.And (all_itp_l),P.True) = TheoremProver.Sat))
    then (failwith ("invalid symm interpolant! (SAT)")) in 
  let _ = Stats.time "process thy clauses" accumulate_theory_clauses tc_l_l in
let _ = M.msg_string M.Normal "Cleaned Interpolants:" in
  let _ = process_interpolants_taut 0 all_itp_l in
  let _ = Stats.time "rename/upd bdds" (folder 0 all_cons) all_itp_l  in
    ((if (!min_i = -1) then (-1) else (!min_i + 1)), !inc_op_l)
  if (Options.getValueOfBool "checktree" 
      && (not (Options.getValueOfBool "dasdill")) 
      && (not (CaddieBdd.bddEqual !curr_bdd2_ref PredTable.bddZero))) then (failwith "Bdd-image interpolant issue (SAT)"); 
  (rsp,!inc_op_l)
  (* INV: !min_i <> -1 -- the if can't fire as foci will have thrown a sat exn *)
}}} *)

let ac_fociModelChecker_refine_trace_abstract reg_array op_array = 
  assert (O.getValueOfBool "focisymm");
  let len_reg_array = Array.length reg_array in 
  let asm_mask_array = Array.create len_reg_array false in 
  let return_reg_array = Array.init len_reg_array (fun i -> reg_array.(i)) in
  let itp_array = Array.create len_reg_array P.True in
  let min_rsp = ref (-1) in
  let ibs_l_ref = ref [] in
  let reg_s_l_ref = ref (Misc.clone "" len_reg_array) in
  let (acc_itp_o,tps_o) =
    if (O.getValueOfBool "lazysymm") then
      let (rsp,_,_,tps_o,acc_itp_o) = ac_block_analyze_trace (Array.to_list reg_array) (Array.to_list op_array) in
      if rsp < 0 then raise TheoremProver.FociSatException else (acc_itp_o,tps_o) 
    else (None,None) in
  let _ = M.msg_string M.Normal "calling build_fmc_constraint_arrays" in
  let (cba,ccba,rma,num_ops,cxca,cxcsl) = build_fmc_constraint_arrays reg_array op_array tps_o in
  let _ = M.msg_string M.Normal "done build_fmc_constraint_arrays" in
  (try
    while ((not(O.getValueOfBool "lazysymm")) && O.getValueOfBool "multistr") || !min_rsp = -1 do
      M.msg_string M.Debug "strengthen again";
      foci_ctr := !foci_ctr + 1;
      let (rsp,inc_op_l) = 
        try refine_trel_symm reg_array cba ccba asm_mask_array rma num_ops acc_itp_o 
        with TheoremProver.FociSatExceptionAsgn plist -> 
            ((if !min_rsp = -1 then 
               reg_s_l_ref := Stats.time "print ctrx" (print_counterexample op_array rma) plist); 
              raise TheoremProver.FociSatException) in
       if (!min_rsp = -1) || (rsp < !min_rsp) then min_rsp := rsp;
       ibs_l_ref := inc_op_l :: !ibs_l_ref; 
       process_inconsistent_ops op_array asm_mask_array (List.rev inc_op_l);
     done;
  with TheoremProver.FociSatException -> ());  (* !min_rsp, !ibs_l_ref,cxcube_s_l)*)
  (!min_rsp, Array.to_list return_reg_array, !ibs_l_ref,!reg_s_l_ref) 
   (* {{{ 
    else if O.getValueOfBool "focisymm" && O.getValueOfBool "lazysymm" then 
        let (rsp,inc_op_l) = (* refine_reach_states *) 
          refine_trel_lazy reg_array op_array asm_mask_array itp_array return_reg_array in
        (min_rsp := rsp; ibs_l_ref := [inc_op_l]) in
        (!min_rsp, Array.to_list return_reg_array, !ibs_l_ref, !reg_s_l_ref)
    else (failwith "calling ac_fociModelChecker_refine_trace_abstract without focisymm!"; assert false)
     * let get_ref_start_pos () = 
       refine_reach_states reg_array op_array conj_block_array
       copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops
       in *)
   (* let refine_trel () =
     refine_trel_states reg_array conj_block_array copy_conj_block_array asm_mask_array itp_array return_reg_array num_ops
   in *)
   (* let refine_trel_sy () = refine_trel_symm reg_array conj_block_array copy_conj_block_array ctrx_cube_array asm_mask_array rev_map_array num_ops 
   in *)
     (* let (ref_start_pos,inc_op_l) = strengthener () in 
       if (Options.getValueOfBool "focisymm") 
          then Stats.time "refine trel symm" refine_trel_sy ()
          else (Stats.time "itp do_refine" get_ref_start_pos ()) in 
         (* if ((fst rv) <> -1 && Options.getValueOfBool "bddpost") 
         then (Stats.time "itp refine trel" refine_trel ());*) 
     (ref_start_pos,inc_op_l) in *)
   (* end strengthen_pass *)
   (* EXN: FociSat }}} *)
    (* with TrelUnsatExn i -> (i, Array.to_list reg_array,[],[]) *)

(** this uses the current transition relation bdd to strengthen the regions
    along the path instead of firing a fresh query to foci *)
let dummy_table = Hashtbl.create 2 (* Sigh. Region.post --- and that damn tar... *)

let fociModelChecker_refine_trace_abstract_abstract
        (reg_array : Region.t array) 
        (op_array : Operation.t array) =
      let len_reg_array = Array.length reg_array in 
      let _ = M.msg_string M.Debug (" len_reg_array = "^(string_of_int len_reg_array)) in
      (* print the bloody thing! *)
      let pr i r_i = M.msg_string M.Debug (Region.to_string r_i) in
      Array.iteri pr reg_array;
        
      assert (len_reg_array = Array.length op_array + 1);
      (* 1. return_reg_array is a copy of reg_array *)
      let return_reg_array = Array.init len_reg_array (fun i -> reg_array.(i)) in
      let back_reg_array = Array.init len_reg_array (fun i -> reg_array.(i)) in
      let rec get_ref_start_pos i =
        if i > 0 then
         begin
           let prereg = pre (back_reg_array.(i)) (op_array.(i-1)) dummy_table in
           let precapreg = (Region.cap (back_reg_array.(i-1)) prereg) in
           let _ = Array.set back_reg_array (i-1) precapreg in
           if Region.is_empty (back_reg_array.(i-1)) then i 
           else get_ref_start_pos (i-1)
         end
        else (-1)
      in
      let rec refine_reg_array i =
        if i > (len_reg_array -1) then () else
          begin
            assert (0 < i && i < len_reg_array); 
            let postreg = post (return_reg_array.(i-1)) (op_array.(i-1)) dummy_table in
            Array.set return_reg_array i postreg;
            refine_reg_array (i+1)
          end   
      in
      (* 2. find the i closest to the error to refine *)
      let ref_start_pos = get_ref_start_pos (len_reg_array - 1) in
      if (ref_start_pos <> -1) then 
      (* 3. strengthen the regions from ref_start_pos to len_reg_array-1 *)  
        refine_reg_array ref_start_pos;
        M.msg_string M.Debug ("refine_abstract_abstract from: "^(string_of_int ref_start_pos));
      (ref_start_pos, Array.to_list return_reg_array)

      
(***********************************************************************************************)
(****************** foci_post_abstract :********************************************************) 
(***********************************************************************************************)

let foci_post_abstract reg op =
  let _ = M.msg_string M.Debug "Foci Post Abstract:" in
  let _ = M.msg_string M.Debug (Operation.to_string op) in
   match reg with
       Region.Atomic atomic_reg ->
	 let (new_location,new_stack) = location_post atomic_reg op  in
	 let post_reg = Region.Atomic { Region.location = new_location; 
					Region.stack = new_stack; 
					Region.data = Region.Abstract 
							{ Region.bdd = PredTable.bddOne;
                                                          Region.lelement = DataLattice.top ;
							  Region.preds = [] };
				      }
	 in
	   (post_reg)
     | _ -> failwith ("Unexpected region in foci_post_abstract!"^(Region.to_string reg))



(* END focimodelchecker ******************************************************************************)
let check_abstract_transition reg op reg' = 
  M.msg_string M.Error "WARNING: in check_abstract_transition -- see code WARNING";
  M.msg_printer M.Debug Region.print reg;
  M.msg_printer M.Debug Operation.print op;
  M.msg_printer M.Debug Region.print reg';
  let b_pred = 
    let p = convert_bdd_to_boolpred (extract_abs_data_reg reg).Region.bdd in
    rename_pred_timeframe 0 p
  in
  let b_pred' = 
    let p =  convert_bdd_to_boolpred (extract_abs_data_reg reg').Region.bdd in
    rename_pred_timeframe 1 p
  in
  let op_cons = 
    let postloc = location_coords_of_region reg' in
    let pred_i_list = (PredTable.get_all_pred_indexes ()) in
    let p = predicate_update_constraints_cached op (postloc,None (* WARNING *)) (pred_i_list,pred_i_list) in
    rename_pred_timeframe 0 p
  in
  try 
    TheoremProver.extract_foci_interpolantN [P.And [b_pred;op_cons];P.Not (b_pred')];
    true
  with e -> 
      begin
      M.msg_string M.Normal "[FAIL] Bad transition!";
      M.msg_string M.Normal ("[FAIL][IN]: "^(P.toString b_pred));
      M.msg_string M.Normal ("[FAIL][OUT]: "^(P.toString b_pred'));
      false
      end



(****************************************************************************)
(******************* Code to combine refinement algorithms ******************)
(****************************************************************************)


(* For -fmc *)
let fociModelChecker_refine_trace_abstract 
        (reg_array : Region.t array)
        (op_array  : Operation.t array) =
   let reg_l = Array.to_list reg_array in
   let op_l = Array.to_list op_array in
   if (O.getValueOfBool "fmcpa") then 
     (PredTable.block_a_t_flag := false;
      print_string "call bat";
      let (rsp,_,_) = block_analyze_trace reg_l op_l [] in
      print_string "done bat";
      (rsp,reg_l))
   else
     try
     let (rsp,ret_reg_l,interesting_block_sets,reg_s_l) = 
       ac_fociModelChecker_refine_trace_abstract reg_array op_array in
     if (rsp <> -1 && !PredTable.block_a_t_flag) then
       let post_reg_s_l = List.map (fun r -> String.concat "," (fst (reg_to_strings r))) (List.tl reg_l) in
       htmlize_trace op_l (List.tl reg_s_l) interesting_block_sets "Weak Transition Relation";
       (rsp, ret_reg_l)
     else if (rsp <> -1 && O.getValueOfBool "localrestart") then
       (print_lp_table ();printTableOfPreds (); dump_abs ();raise NoNewPredicatesException)
     else if (rsp <> -1 && O.getValueOfBool "lazysymm") then
       (O.setValueOfBool "localrestart" true; dump_abs (); (rsp,ret_reg_l)) 
     else if (rsp = -1 && O.getValueOfBool "lazysymm") then 
       (rsp, ret_reg_l)
     else
       let _ = assert (List.length reg_s_l <> List.length reg_l) in
       let reg_l = Array.to_list reg_array in
       let op_l = Array.to_list op_array in
       let _ = PredTable.block_a_t_flag := false in
       let (rsp,_,_) = block_analyze_trace reg_l op_l [] in
       if (rsp <> -1 && not !PredTable.block_a_t_flag) then raise NoNewPredicatesException
       else if (rsp <> -1 && Options.getValueOfBool "fmc" && 
                not (Options.getValueOfBool "hybrid")) 
            then 
              let (rsp,ret_reg_l,_,_) = ac_fociModelChecker_refine_trace_abstract reg_array op_array in 
                (rsp,ret_reg_l)
            else (rsp,ret_reg_l)  
    with TheoremProver.FociSatException -> (-1,[])



     
(* WARNING: Does not work for TAR/ETAR *)
(* For vanilla blast *)
(* Could also call the real block_at ===> ac_block_at  and call this below function block_a_t *)
(* could also flip the order below -- if block_at fails && -focisymm is on, try fmc-refine ...*)
let trel_refine_block_analyze_trace reg_l op_l id_l =
  if (Options.getValueOfBool "focisymm") then 
    let (rsp,ret_reg_l) = fociModelChecker_refine_trace_abstract (Array.of_list reg_l) (Array.of_list op_l) in
    if (rsp >= 0) then (rsp,(List.nth ret_reg_l rsp), true)  else (-1,Region.bot,false)
  else (block_analyze_trace reg_l op_l id_l)
    
      
