(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

(* $Id: abstraction.mli,v 1.5 2007/08/29 01:41:53 beyer Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module implements the abstraction interface
 * connecting the model checker to the concrete transition system.
 * It defines the abstract pre and post, and various routines for
 * focus, shrink, etc used by the model checker.
 *) 


open BlastArch


module Make_C_Abstraction :
  functor(C_SD : SYSTEM_DESCRIPTION with module Command = BlastCSystemDescr.C_Command) ->
  sig
    include ABSTRACTION
    (* val build_custom_abstraction : (((string * ((Ast.Expression.expression * bool) list)) list) 
				      * ((Ast.Expression.expression * (Ast.Expression.expression * bool) list) list option) 
				      * ((Ast.Expression.expression*string) list))-> unit
    val initKill : C_SD.location option -> Ast.Predicate.predicate list -> unit
*)	
    (* create a region: takes a location, a flag that sets the stack (flag = true -> EveryStack, flag = false -> [])
	the set of seed predicates, the initial state and returns a region
	*)
      (** As the name suggests, this sets up the abstraction by loading in predicates and such *)
    val initialize_abstraction : unit -> unit
    val dump_abs : unit -> unit
    val save_region_abs : Region.t -> unit

    val ac_create_region : bool -> C_SD.location list -> bool -> Ast.Predicate.predicate -> Ast.Predicate.predicate -> Region.t

    val create_region : C_SD.location list -> bool -> Ast.Predicate.predicate -> Ast.Predicate.predicate -> Region.t

    val print_stats : Format.formatter -> unit -> unit
    val check_races : Ast.Expression.lval option ref
    val assumed_invariants : Ast.Predicate.predicate list ref
    val assumed_invariants_region : Region.t ref
    val assumed_invariants_list_to_region : unit -> Region.t

  end with type Operation.t = C_SD.edge
       and type Operation.exp = Ast.Expression.expression


