(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

(* $Id: events.ml,v 1.7 2007/08/29 01:41:53 beyer Exp $
 *
 * This file is part of the BLAST Project.
 *)
open BlastArch

module E = Ast.Expression
module P = Ast.Predicate
module M = Message

module Make_Events = 
  functor(PredTable : PREDTABLE with module C_SD.Command = BlastCSystemDescr.C_Command) ->
struct
  module C_SD = PredTable.C_SD
  module Command = PredTable.C_SD.Command
  type eventHandler = string * int list
  
  let eventHandler_to_string (s,is) = "Event: "^s^" : "^(Misc.string_of_int_list is)
  
  let int_to_expr i = Ast.Expression.Constant (Ast.Constant.Int i)  
    
  let expr_to_int e = 
    match e with 
      Ast.Expression.Constant (Ast.Constant.Int i) -> i
    | _ -> (assert false; failwith "bad call to expr_to_int")

  let eventHandler_to_command (s,l) = 
    {Command.code = Command.FunctionCall (E.FunctionCall(E.Lval(E.Symbol(s)),List.map int_to_expr l));
     Command.read = [];
     Command.written = [];} 

  let _plug_args op args =
    let cmd' = 
      match (C_SD.get_command op).Command.code with
        Command.FunctionCall (E.FunctionCall(E.Lval(E.Symbol s) ,_)) 
      | Command.FunctionCall (E.Assignment (_,_,E.FunctionCall (E.Lval (E.Symbol s), _))) ->
          {Command.code = Command.FunctionCall (E.FunctionCall(E.Lval(E.Symbol s), args));
           Command.read = [];
           Command.written = [];}
      | _ -> (assert false;failwith "error in Events._plug_args") in
    C_SD.make_edge (C_SD.get_source op) (C_SD.get_target op) cmd' 

  let lit_to_int (i,b) = 2*i + (if b then 1 else 0)
  
  let int_to_lit n = 
    let rv = (n / 2, n mod 2 <> 0) in
    assert (n = (lit_to_int rv));
    rv

  let get_lit_bdd (i,b) = 
    let bddi = PredTable.getPredIndexBdd i in
    if b then bddi else CaddieBdd.bddNot bddi

  let deconstructAsyncCall op =
    M.msg_string M.Debug ("in deconstructAsyncCall:"^(C_SD.edge_to_string op));
    match (C_SD.get_command op).Command.code with
      Command.FunctionCall (E.FunctionCall (E.Lval (E.Symbol s), args)) 
    | Command.FunctionCall (E.Assignment (_,_,E.FunctionCall (E.Lval (E.Symbol s), args))) ->
        let is = List.map expr_to_int args in
        let ls = List.map int_to_lit is in
        let bdd = List.fold_left (fun b l -> CaddieBdd.bddAnd b (get_lit_bdd l)) PredTable.bddOne ls in
          ((s,is),bdd)
    | _ -> (assert false; failwith "error in Events.deconstructAsyncCall")

      
  let constructAsyncCall (op,bdd) =
    M.msg_string M.Debug (Printf.sprintf "in constructAsyncCall: %s : %s" (C_SD.edge_to_string op) (P.toString (PredTable.convertBddToPred bdd)));
    let lits = PredTable.lits_of_cube (PredTable.bdd_cube_cover bdd) in
    let lits' = List.filter (fun (i,_) -> not (PredTable.is_pred_global i)) lits in
    let is = List.map lit_to_int lits' in
    let rv = _plug_args op (List.map int_to_expr is) in
    let _ = deconstructAsyncCall rv in
    rv

end


