(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)
module E :
  sig
    type binary_op
    type lval
    type expression
  end
module P :
  sig
    type predicate = Ast.Predicate.predicate 
  end
val convertToVampyreSyntax : P.predicate -> Input.pred 
val set_simplify_query_flag : unit -> unit
val kill_simplify_server : unit -> unit
val ques_flag : bool ref
val queryExp_proof : P.predicate -> Proof.proof option
val uclidSat : P.predicate -> string list
val uclid_output : string -> P.predicate -> string -> unit
val getUsefulPredicates : P.predicate -> P.predicate list
val stats_nb_foci_queries : int ref
exception FociSatException
exception FociSatExceptionAsgn of string list
type foci_return = Sat | Unsat of P.predicate
val extract_foci_preds :
  P.predicate list * P.predicate list -> P.predicate list
val extract_foci_preds_parity :
  P.predicate list * P.predicate list -> (P.predicate * bool) list
val extract_foci_interpolant :
  P.predicate list * P.predicate list -> P.predicate
val interpolant_sat : P.predicate * P.predicate -> foci_return
val interpolant_unsat : P.predicate -> bool
val extract_foci_interpolantN : P.predicate list -> P.predicate list
val extract_foci_interpolantN_noconv : P.predicate list -> P.predicate list
val extract_foci_interpolantN_symm :
  P.predicate list -> P.predicate list * 'a list
val extract_minp_interpolantN : P.predicate list -> P.predicate list
val extract_foci_interpolantN_dasdill :
  P.predicate list -> P.predicate list -> P.predicate list
val foci_inc_pop : unit -> unit
val foci_inc_push : P.predicate -> unit
val foci_inc_check_sat : unit -> bool
val foci_inc_check_unsat : P.predicate -> bool
val foci_inc_is_contra : unit -> bool
val foci_inc_reset : unit -> unit
val queryExpContext : P.predicate -> P.predicate -> bool
val queryExp : P.predicate -> bool
val block_assert : P.predicate -> unit
val block_pop : unit -> unit
val block_reset : unit -> unit
val _is_contra : unit -> bool
val check_unsat : P.predicate -> bool
val _assume : P.predicate -> unit
val _pop : unit -> unit
val generateTest : 'a -> 'b
