(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)
type ('a, 'b, 'c) automaton = {
  aut_id : int;
  mutable start_location : ('a, 'b, 'c) location option;
  mutable end_location : ('a, 'b, 'c) location option;
  mutable aut_attributes : 'a;
}
and ('a, 'b, 'c) location = {
  loc_id : int;
  mutable succs : ('c * ('a, 'b, 'c) location) list;
  mutable preds : ('c * ('a, 'b, 'c) location) list;
  automaton : ('a, 'b, 'c) automaton;
  mutable loc_attributes : 'b;
}

val location_function_name : ('a, 'b, 'c) location -> string
val create_location :
  string -> ('a, 'b, 'c) automaton -> 'b -> ('a, 'b, 'c) location
val create_automaton : 'a -> ('a, 'b, 'c) automaton
val get_start_location : ('a, 'b, 'c) automaton -> ('a, 'b, 'c) location
val get_end_location : ('a, 'b, 'c) automaton -> ('a, 'b, 'c) location
val get_automaton_attributes : ('a, 'b, 'c) automaton -> 'a
val set_start_location :
  ('a, 'b, 'c) automaton -> ('a, 'b, 'c) location -> unit
val set_end_location :
  ('a, 'b, 'c) automaton -> ('a, 'b, 'c) location -> unit
val get_succs : ('a, 'b, 'c) location -> ('c * ('a, 'b, 'c) location) list
val get_preds : ('a, 'b, 'c) location -> ('c * ('a, 'b, 'c) location) list
val get_location_attributes : ('a, 'b, 'c) location -> 'b

val add_edge : ('a, 'b, 'c) location -> ('a, 'b, 'c) location -> 'c -> unit

val remove_edge :
  ('a, 'b, 'c) location -> ('a, 'b, 'c) location -> 'c -> unit
val set_location_attributes : ('a, 'b, 'c) location -> 'b -> unit
