(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

{
module E = VampyreErrormsg
open Inputparse

let lexerror msg lexbuf = 
  E.s (E.unimp "%s" msg)

}

let digit    = ['0'-'9']
let letdig   = ['0'-'9' 'a'-'z' 'A'-'Z' '_' '@' '#']
let othersyms =[ '-' '$' '#' '!' '+' '=' 
                '<' '>' ',' '?' '\'']
let alphlet  = ['A'-'Z' 'a'-'z' '_' '$' '@']
let capital  = ['A'-'Z']
let small    = ['a'-'z' '$' '_']
let ws       = [' ' '\009' '\012']
let pathname = ['a'-'z' 'A'-'Z' '0'-'9' '.' '/' '\\' '-']

rule token = parse
    '\r'                { token lexbuf}
  | '\n'		{ begin
			    E.startNewline (Lexing.lexeme_end lexbuf);
			    token lexbuf 
			  end }
  | "//"[^'!''\n']*'\n'
                        { begin
                            E.startNewline (Lexing.lexeme_end lexbuf);
			    token lexbuf
                          end }
  | "CONSTANT"            { CONSTANT }
  | "sizeof"            { SIZEOF }
  | "int"               { INT }
  | "char"              { CHAR }
  | "long"              { LONG }
  | "float"             { FLOAT }
  | "double"            { DOUBLE }
  | "void"              { VOID }
  | ws			{ token lexbuf }
  | '['                 { LINDEX }
  | ']'                 { RINDEX }
  | '('			{ LPAREN }
  | ')'			{ RPAREN }
  | "->"		{ ARROW }
  | "->*"		{ ARROWSTAR }
  | "++"                { ICR }
  | "--"                { DECR }
  | '.'			{ DOT }
  | ".*"		{ DOTSTAR }
  | '+'			{ PLUS }
  | '-'			{ MINUS }
  | '*'			{ MUL }
  | '?'                 { QUESTION }
  | '&'                 { BITAND }
  | "&&"                { AND }
  | '|'                 { BITOR  }
  | "||"                { OR  }
  | '^'                 { XOR  }
  | '/'                 { DIV  }
  | '!'                 { NOT  }
  | '~'                 { BITNOT  }
  | '%'                 { REM  }
  | ','                 { COMMA  }
  | ';'                 { SEMICOLON  }
  | ':'                 { COLON  }
  | "<<"		{ LSHIFT }
  | ">>"		{ RSHIFT }
  | "<"			{ LT }
  | ">"			{ GT }
  | "<="		{ LE }
  | ">="		{ GE }
  | "=="		{ EQ }
  | "!="		{ NE }
  | "="                 { ASSIGN }
  | "*="                { MULTassign }
  | "/="                { DIVassign }
  | "%="                { MODassign }
  | "+="                { PLUSassign }
  | "-="                { MINUSassign }
  | "<<="                { LSassign }
  | ">>="                { RSassign }
  | "&="                { ANDassign }
  | "^="                { XORassign }
  | "|="                { ORassign }
  | "[]"                { INDEX }
  | "&&"		{ AND }
  | "||"		{ OR }
  | (digit)+	{ let str = Lexing.lexeme lexbuf in
			  let len = String.length str in
			  let zero = Char.code '0' in
			  let rec accum a d =
			    let acc c = a + (d * ((Char.code c) - zero)) in
			    function
			      0 -> let c = str.[0] in
				   if c='~' then - a else (acc c)
			    | i -> accum (acc str.[i]) (d * 10) (i - 1)
			  in
			  Num (accum 0 1 (len-1)) }
  | (alphlet)letdig*	{ Id    (Lexing.lexeme lexbuf) }

  | eof			{ EOF }
  | _			{ 
                          begin
                            lexerror ("Illegal Character '" ^ 
                                      (Lexing.lexeme lexbuf)) lexbuf;
			    token lexbuf
			  end }

