(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

open Ast

let constructInitialPredicates predFile =
  let readAndAddPredicates filename =
    let doParse inchan =
      let lexbuf = Lexing.from_channel inchan in
      Inputparse.main Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in seed predicates...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
	ignore (Message.msg_string Message.Normal "Seed predicates read.\n") ;	
	List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
	Message.msg_string Message.Normal ("Read "^(string_of_int (List.length a))^" predicates");
	a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find predicate information in "^filename^".\n"); [])
    | e -> (Message.msg_string Message.Error (Printexc.to_string e); [])
  in
  if (predFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddPredicates predFile

