(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)
if Array.length Sys.argv < 2 then
  raise (Failure ("Missing parameters. Call with: " ^ Sys.argv.(0) ^ " <spec file> [<C files>]\nIf no C file is supplied, all .c files in the cwd will be checked."));

let specFile = Sys.argv.(1) in
print_string "Parsing sources....\n";
flush stdout;
let fnames =
  if Array.length Sys.argv >= 3 then
    Array.to_list (Array.sub Sys.argv 2 (Array.length Sys.argv - 2))
  else
    (let dir = Unix.opendir "." in
    let rec getFilenames () =
      try
	let fname = Unix.readdir dir in
	if (String.length fname >= 2 && compare ".c" (String.sub fname (String.length fname - 2) 2) == 0)
	    && compare "merged.c" fname != 0
	    && compare "instrumented.c" fname != 0 then
	  fname :: getFilenames ()
	else
	  getFilenames ()
      with End_of_file -> [] in
    let fnames = getFilenames () in
    Unix.closedir dir;
    fnames) in
let files = List.map Common.parseFile fnames in
let file = Mergecil.merge files "merged.c" in

let merged = open_out "merged.c" in
Common.outputFile merged file;
close_out merged;

(*print_string "Initial code:\n";
Common.printFile file;*)

Rmtmps.removeUnusedTemps file;

let spec = Spec.parseSpec specFile in

print_string "Spec read\n";
flush stdout;

(*print_string "\nSpec:\n";
Spec.printSpec spec;*)

let file = Spec.instrument spec file in

(*print_string "\Before rmtmps:\n";
Common.printFile file;

Rmtmps.removeUnusedTemps file;*)

(*print_string "\nInstrumented code:\n";
Common.printFile file;*)

let outf = open_out "instrumented.c" in
Common.outputFile outf file;
close_out outf;

print_string "Done\n"
