(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* This defines the signature that a satisfiability procedure must implement *)

exception Contradiction of Proof.proof   (* a proof of False *)

type msgbody =
    Equality of Dag.node * Dag.node
  | Literal of Dag.node


type ident = string                     (* The unique identifier of a 
                                         * satsifiability procedure *)

type message = 
    { body    : msgbody;							  (* literal or equality *)
      owner   : ident;                  (* Who is responsible for producing a 
                                         * proof of this  *)
      id      : int;                    (* The responsible satisf. proc can 
                                         * use this to later identify the 
                                         * message  *)
      goal    : bool;                   (* If it is a goal. In this case the 
                                         * asserted predicate is the negation 
                                         * of "literal"  *)
    } 
      

                                        (* There are the entry points in a 
                                         * decision procedure *)
type entries = 
                                        (* Record a snapshot *)
    { snapshot  : unit -> unit;
                                        (* Discard all assertions since last 
                                         * undone snapshot *)
      undo      : unit -> unit;
                                        (* Assert a literal (the second 
                                         * argument). Accumulate the list of 
                                         * equalities to the first argument  *)
	assume    : message list -> message -> message list;
                                        (* Prove a message generated earlier.*)
	proveIt   : message -> Proof.proof;

        
	debug     : bool ref;             (* For setting a debug flag *)

	ident     : ident;                (* The identifier *)
	
	reinitialize : unit -> unit;    (* reinitializes the decision procedure -- by resetting values of all ref'fed, mutable objects *)  
    } 


                               (* There is a global pointer to a procedure 
                                * for proof generation. satisfiability 
                                * procedures can call this when they need to 
                                * produce a proof for a message belonging to 
                                * some _OTHER_ satproc.  *)

(* [Jhala]: I need this silly thing ... *)
(*let rec dummyNode = {Dag.id = 0;Dag.op = Dag.OpName "";Dag.args =[];Dag.root = dummyNode; Dag.eqClass =[]; Dag.use = []; Dag.isAsserted = false; Dag.mesg = dummyMesg}
 and dummyMesg = {body = Literal dummyNode;owner ="";id = 0; goal = false}
*)
      
let proveIt : (message -> Proof.proof) ref = 
  ref (fun _ -> VampyreErrormsg.s (VampyreErrormsg.bug "proveIt is not defined"))

    

(* VampyrePretty print a message *)
(* 
 * [weimer] We print out the dag node, the (sender,id) pair and whether or
 * not the message is a goal. 
 *)
let dm () msg = 
  match msg.body with
    Equality(dn1,dn2) -> 
      VampyrePretty.dprintf "{(%a == %a) (%s,%d)%s}" Dag.dn dn1 Dag.dn dn2 
	msg.owner msg.id (if msg.goal then " GOAL" else "")
  | Literal(dn1) -> 
      VampyrePretty.dprintf "{%a (%s,%d)%s}" Dag.dn dn1 
	msg.owner msg.id (if msg.goal then " GOAL" else "")
	
module type SATPROC = 
  sig
  
                                        (* Called to initialize the decision 
                                         * procedure and to create a table of 
                                         * the entry points. *)
    val initialize : ident -> entries
  end
